######################################################################
# 2D Arm Manipulator
#
#  shoulder    elbow      wrist
#
#     |_ _______ |_ _______ |_ ____
#
#        humerus     ulna      hand
#
# All values in cm's or degrees
#
# set arm(humerus) 28
# set arm(ulna) 28
# set arm(hand) 18
# set arm(shoulder) 0
# set arm(shoulder_range) { -90 110 }
# set arm(elbow) 0
# set arm(elbow_range) { 0 150 }
# set arm(wrist) 0
# set arm(wrist_range) { -90 90 }

source vector.tcl

proc CreateArm {} {

    global arm

    set arm(humerus) 28
    set arm(ulna) 28
    set arm(hand) 18
    
    set arm(shoulder) 0
    set arm(shoulder_origin) { 0 0 0 1 }
    set arm(shoulder_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set arm(shoulder_range) { -90 110 }
    set arm(elbow) 0
    set arm(elbow_origin) { 0 0 0 1 }
    set arm(elbow_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set arm(elbow_range) { 0 165 }
    set arm(wrist) 0
    set arm(wrist_origin) { 0 0 0 1 }
    set arm(wrist_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set arm(wrist_range) { -90 90 }
}

proc CreateLegs {} {

    global legLeft

    set legLeft(humerus) 50
    set legLeft(ulna) 45
    set legLeft(hand) 22
    
    set legLeft(shoulder) 0
    set legLeft(shoulder_origin) { 0 0 0 1 }
    set legLeft(shoulder_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legLeft(shoulder_range) { -30 150 }
    set legLeft(elbow) 0
    set legLeft(elbow_origin) { 0 0 0 1 }
    set legLeft(elbow_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legLeft(elbow_range) { -150 0 }
    set legLeft(wrist) 0
    set legLeft(wrist_origin) { 0 0 0 1 }
    set legLeft(wrist_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legLeft(wrist_range) { 0 110 }

    global legRight

    set legRight(humerus) 50
    set legRight(ulna) 45
    set legRight(hand) 22
    
    set legRight(shoulder) 0
    set legRight(shoulder_origin) { 0 0 0 1 }
    set legRight(shoulder_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legRight(shoulder_range) { -30 150 }
    set legRight(elbow) 0
    set legRight(elbow_origin) { 0 0 0 1 }
    set legRight(elbow_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legRight(elbow_range) { -150 0 }
    set legRight(wrist) 0
    set legRight(wrist_origin) { 0 0 0 1 }
    set legRight(wrist_twist) { { 0 0 0 0 } { 0 0 1 0 } }
    set legRight(wrist_range) { 0 110 }
}

######################################################################
# Solves Inverse Kinematics Subproblem 1
#   Find the rotation angle which brings one point to another
#
# Returns the Rotation Angle
proc Subproblem1 { twist ptR ptP ptQ } {

    set w [lindex $twist 1]

    set u [vexpr $ptP - $ptR]
    set v [vexpr $ptQ - $ptR]
    
    ##################################################
    # Get the projection of the vectors in the plane perpendicular the axis
    #
    set u [vexpr $u - $w * [VdotV $w $u]] 
    set v [vexpr $v - $w * [VdotV $w $v]] 

    #set v [Vnormalize $v]
    #set u [Vnormalize $u]
    
    set y [VdotV $w [VcrossV $u $v]]
    set x [VdotV $u $v]

    return [expr atan2($y, $x)]
}

######################################################################
# Solves Inverse Kinematics Subproblem 3
#   Find the rotation angle which brings one point to another
#
# Returns the Rotation Angle
proc Subproblem3 { ptQ twist ptR ptP fDelta2 } {

#puts "Subproblem 3:"
#puts "  ptQ = $ptQ"
#puts "  twist = $twist"
#puts "  ptR = $ptR"
#puts "  ptP = $ptP"
#puts "  fDelta2 = $fDelta2"

    set w [lindex $twist 1]
#puts "  w = $w"    

    set u [vexpr $ptP - $ptR]
    set v [vexpr $ptQ - $ptR]
#puts "  u = $u"    
#puts "  v = $v"    

    set fDeltaP2 [expr $fDelta2 - [VdotV $w [vexpr $ptP - $ptQ]]]
#puts "  fDeltaP2 = $fDeltaP2"

    ##################################################
    # Get the projection of the vectors in the plane perpendicular the axis
    #
    set u [vexpr $u - $w * [VdotV $w $u]] 
    set v [vexpr $v - $w * [VdotV $w $v]] 
#puts "  u = $u"    
#puts "  v = $v"    

    #set v [Vnormalize $v]
    #set u [Vnormalize $u]
    
    set y [VdotV $w [VcrossV $u $v]]
    set x [VdotV $u $v]

#puts "Subproblem 3: ($x, $y)"

    set fTheta0 [expr atan2($y, $x)] 
#puts "fTheta0 = $fTheta0"

    set num [expr [VdotV $u $u] + [VdotV $v $v] - $fDeltaP2]
    set den [expr 2.0*[Vmagnitude $u]*[Vmagnitude $v]]
#puts "num = $num"
#puts "den = $den"
    if { $den == 0.0 } {
	puts "ERROR: Subproblem 3: den == 0"
	set fTheta1 0.0
    } else {
	set fCos [Max [Min [expr $num / $den] 1.0] -1.0]
#puts "fCos = $fCos"
	set fTheta1 [expr acos($fCos)]
    }

#puts "fTheta1 = $fTheta1"

    return [list [expr $fTheta0 + $fTheta1] [expr $fTheta0 - $fTheta1]]
}

proc CalculateArmJointAngles { pArm mGs_w } {

    global $pArm

    ##################################################
    # Update Manipulator Twists in case Linkage Lengths changed
    #
    
    # Shoulder is constant
#    puts "Shoulder Origin = [subst $[subst $pArm](shoulder_origin)]"
#    puts "Shoulder Twist = [subst $[subst $pArm](shoulder_twist)]"

    # Elbow
    set [subst $pArm](elbow_origin) [list [subst $[subst $pArm](humerus)] 0 0 1]
    set v [VcrossV [vexpr [lindex [subst $[subst $pArm](elbow_twist)] 1] * -1] [subst $[subst $pArm](elbow_origin)]]
    set [subst $pArm](elbow_twist) [list $v [lindex [subst $[subst $pArm](elbow_twist)] 1]]
#    puts "Elbow Origin = [subst $[subst $pArm](elbow_origin)]"
#    puts "Elbow Twist = [subst $[subst $pArm](elbow_twist)]"

    # Wrist
    set [subst $pArm](wrist_origin) [list [expr [subst $[subst $pArm](humerus)] + [subst $[subst $pArm](ulna)]] 0 0 1]
    set v [VcrossV [vexpr [lindex [subst $[subst $pArm](wrist_twist)] 1] * -1] [subst $[subst $pArm](wrist_origin)]]
    set [subst $pArm](wrist_twist) [list $v [lindex [subst $[subst $pArm](wrist_twist)] 1]]
#    puts "Wrist Origin = [subst $[subst $pArm](wrist_origin)]"
#    puts "Wrist Twist = [subst $[subst $pArm](wrist_twist)]"

    ##################################################
    # Solve for Joint Angles
    #

    set ptOrigin { 0 0 0 1 }
    set ptShoulderS $ptOrigin
    set ptHandS [MV $mGs_w $ptOrigin]

    ##################################################
    # Subproblem 3 -- Elbow
    #

    set vDelta [vexpr $ptHandS - $ptShoulderS]
    set fDelta [VdotV $vDelta $vDelta]; # This is the distance squared

    set pfAngle [Subproblem3 [subst $[subst $pArm](shoulder_origin)] [subst $[subst $pArm](elbow_twist)] [subst $[subst $pArm](elbow_origin)] [subst $[subst $pArm](wrist_origin)] $fDelta]   

    set ppfJoints ""
    foreach fAngle $pfAngle {
	set fElbow [RadToDeg $fAngle]
	
	if { [lindex [subst $[subst $pArm](elbow_range)] 0] <= $fElbow && $fElbow <= [lindex [subst $[subst $pArm](elbow_range)] 1] } {
	    lappend ppfJoints [list 0 $fElbow]
	} else {
	    lappend ppfJoints [list 1 [Min [Max $fElbow [lindex [subst $[subst $pArm](elbow_range)] 0]] [lindex [subst $[subst $pArm](elbow_range)] 1]]]
	}
    }

#puts "Joints = $ppfJoints"

    ##################################################
    # Subproblem 1 -- Shoulder
    #

    set ppfJointsElbow $ppfJoints
    set ppfJoints ""

    foreach pfJoints $ppfJointsElbow {
	set ptHandE [MV [Mtwist [subst $[subst $pArm](elbow_twist)] [DegToRad [lindex $pfJoints 1]]] [subst $[subst $pArm](wrist_origin)]]
	set fShoulder [RadToDeg [Subproblem1 [subst $[subst $pArm](shoulder_twist)] [subst $[subst $pArm](shoulder_origin)] $ptHandE $ptHandS]]
	
	if { [lindex [subst $[subst $pArm](shoulder_range)] 0] <= $fShoulder && $fShoulder <= [lindex [subst $[subst $pArm](shoulder_range)] 1] } {
	    lappend ppfJoints [concat $pfJoints $fShoulder]
	} else {
	    lappend ppfJoints [concat [expr 1 + [lindex $pfJoints 0]] [lrange $pfJoints 1 1] [Min [Max $fShoulder [lindex [subst $[subst $pArm](shoulder_range)] 0]] [lindex [subst $[subst $pArm](shoulder_range)] 1]]]
	}
    }

#puts "Joints = $ppfJoints"

    ##################################################
    # Subproblem 1 -- Wrist
    #

    set ppfJointsShoulder $ppfJoints
    set ppfJoints ""

    foreach pfJoints $ppfJointsShoulder {
	set ptP [vexpr $ptOrigin + $ptOrigin - [subst $[subst $pArm](wrist_origin)]]

	set ptQ [MV [Mtwist [subst $[subst $pArm](elbow_twist)] [expr -[DegToRad [lindex $pfJoints 1]]]] [MV [Mtwist [subst $[subst $pArm](shoulder_twist)] [expr -[DegToRad [lindex $pfJoints 2]]]] [MV $mGs_w $ptP]]]
	set fWrist [RadToDeg [Subproblem1 [subst $[subst $pArm](wrist_twist)] [subst $[subst $pArm](wrist_origin)] $ptP $ptQ]]
	
	if { [lindex [subst $[subst $pArm](wrist_range)] 0] <= $fWrist && $fWrist <= [lindex [subst $[subst $pArm](wrist_range)] 1] } {
	    lappend ppfJoints [concat $pfJoints $fWrist]
	} else {
	    lappend ppfJoints [concat [expr 1 + [lindex $pfJoints 0]] [lrange $pfJoints 1 2] [Min [Max $fWrist [lindex [subst $[subst $pArm](wrist_range)] 0]] [lindex [subst $[subst $pArm](wrist_range)] 1]]]
	}
    }

#puts "Joints = $ppfJoints"

#puts "Store the Joint Angles"
#puts "Shoulder Angle = $pfShoulder"
#puts "Elbow Angle = $pfElbow"

    ##################################################
    # Store the Joint Angles
    #

    set pfJointsBest [lindex $ppfJoints 0]

    foreach pfJoints $ppfJoints {
	if { [lindex $pfJoints 0] < [lindex $pfJointsBest 0] } {
	    set pfJointsBest $pfJoints
	}
    }

    set [subst $pArm](shoulder) [lindex $pfJointsBest 2]
    set [subst $pArm](elbow) [lindex $pfJointsBest 1]
    set [subst $pArm](wrist) [lindex $pfJointsBest 3]
}

set mShoulder { 1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1 }
set mHand { 1 0 0 0 0 1 0 0 0 0 1 0 20 0 0 1 }


